@@ -2,11 +2,13 @@ import torch
 from torch import nn, einsum, broadcast_tensors
 import torch.nn.functional as F
 
-from einops import rearrange
+from einops import rearrange, repeat
 from einops.layers.torch import Rearrange
 
 from torch_geometric.nn import MessagePassing
+
 # types
+
 from typing import Optional, List, Union
 from torch_geometric.typing import Adj, Size, OptTensor, Tensor
 
@@ -15,6 +17,11 @@ from torch_geometric.typing import Adj, Size, OptTensor, Tensor
 def exists(val):
     return val is not None
 
+def safe_div(num, den, eps = 1e-8):
+    res = num.div(den.clamp(min = eps))
+    res.masked_fill_(den == 0, 0.)
+    return res
+
 def batched_index_select(values, indices, dim = 1):
     value_dims = values.shape[(dim + 1):]
     values_shape, indices_shape = map(lambda t: list(t.shape), (values, indices))
@@ -40,6 +47,17 @@ def fourier_encode_dist(x, num_encodings = 4, include_self = True):
     x = torch.cat((x, orig_x), dim = -1) if include_self else x
     return x
 
+def embedd_token(x, dims, layers):
+    stop_concat = -len(dims)
+    to_embedd = x[:, stop_concat:].long()
+    for i,emb_layer in enumerate(layer):
+        # the portion corresponding to `to_embedd` part gets dropped
+        x = torch.cat([ x[:, :stop_concat], 
+                        emb_layer( to_embedd[:, i] ) 
+                      ], dim=-1)
+        stop_concat = x.shape[-1]
+    return x
+
 # swish activation fallback
 
 class Swish_(nn.Module):
@@ -48,6 +66,23 @@ class Swish_(nn.Module):
 
 SiLU = nn.SiLU if hasattr(nn, 'SiLU') else Swish_
 
+# helper classes
+
+# this follows the same strategy for normalization as done in SE3 Transformers
+# https://github.com/lucidrains/se3-transformer-pytorch/blob/main/se3_transformer_pytorch/se3_transformer_pytorch.py#L95
+
+class CoorsNorm(nn.Module):
+    def __init__(self, eps = 1e-8):
+        super().__init__()
+        self.eps = eps
+        self.fn = nn.Sequential(nn.LayerNorm(1), nn.GELU())
+
+    def forward(self, coors):
+        norm = coors.norm(dim = -1, keepdim = True)
+        normed_coors = coors / norm.clamp(min = self.eps)
+        phase = self.fn(norm)
+        return (phase * normed_coors)
+
 # classes
 
 class EGNN(nn.Module):
@@ -63,10 +98,15 @@ class EGNN(nn.Module):
         dropout = 0.0,
         init_eps = 1e-3,
         norm_feats = False,
+        norm_coors = False,
         update_feats = True,
-        update_coors = True
+        update_coors = True,
+        only_sparse_neighbors = False,
+        valid_radius = float('inf'),
+        m_pool_method = 'sum'
     ):
         super().__init__()
+        assert m_pool_method in {'sum', 'mean'}, 'pool method must be either sum or mean'
         assert update_feats or update_coors, 'you must update either features, coordinates, or both'
 
         self.fourier_features = fourier_features
@@ -83,6 +123,9 @@ class EGNN(nn.Module):
         )
 
         self.node_norm = nn.LayerNorm(dim) if norm_feats else nn.Identity()
+        self.coors_norm = CoorsNorm() if norm_coors else nn.Identity()
+
+        self.m_pool_method = m_pool_method
 
         self.node_mlp = nn.Sequential(
             nn.Linear(dim + m_dim, dim * 2),
@@ -105,6 +148,8 @@ class EGNN(nn.Module):
         ) if update_coors else None
 
         self.num_nearest_neighbors = num_nearest_neighbors
+        self.only_sparse_neighbors = only_sparse_neighbors
+        self.valid_radius = valid_radius
 
         self.init_eps = init_eps
         self.apply(self.init_)
@@ -114,8 +159,12 @@ class EGNN(nn.Module):
             # seems to be needed to keep the network from exploding to NaN with greater depths
             nn.init.normal_(module.weight, std = self.init_eps)
 
-    def forward(self, feats, coors, edges = None, mask = None):
-        b, n, d, fourier_features, num_nearest = *feats.shape, self.fourier_features, self.num_nearest_neighbors
+    def forward(self, feats, coors, edges = None, mask = None, adj_mat = None):
+        b, n, d, device, fourier_features, num_nearest, valid_radius, only_sparse_neighbors = *feats.shape, feats.device, self.fourier_features, self.num_nearest_neighbors, self.valid_radius, self.only_sparse_neighbors
+
+        if exists(mask):
+            num_nodes = mask.sum(dim = -1)
+
         use_nearest = num_nearest > 0
 
         rel_coors = rearrange(coors, 'b i d -> b i () d') - rearrange(coors, 'b j d -> b () j d')
@@ -124,7 +173,30 @@ class EGNN(nn.Module):
         i = j = n
 
         if use_nearest:
-            nbhd_indices = rel_dist[..., 0].topk(num_nearest, dim = -1, largest = False).indices
+            ranking = rel_dist[..., 0]
+
+            if exists(mask):
+                rank_mask = mask[:, None, :] * mask[:, None, :]
+                ranking.masked_fill_(~rank_mask, 1e5)
+
+            if exists(adj_mat):
+                if len(adj_mat.shape) == 2:
+                    adj_mat = repeat(adj_mat, 'i j -> b i j', b = b)
+
+                if only_sparse_neighbors:
+                    num_nearest = int(adj_mat.float().sum(dim = -1).max().item())
+                    valid_radius = 0
+
+                self_mask = rearrange(torch.eye(n, device = device, dtype = torch.bool), 'i j -> () i j')
+
+                ranking.masked_fill_(self_mask, -1.)
+                adj_mat.masked_fill_(self_mask, False)
+                ranking.masked_fill_(adj_mat, 0.)
+
+            nbhd_ranking, nbhd_indices = ranking.topk(num_nearest, dim = -1, largest = False)
+
+            nbhd_mask = nbhd_ranking <= valid_radius
+
             rel_coors = batched_index_select(rel_coors, nbhd_indices, dim = 2)
             rel_dist = batched_index_select(rel_dist, nbhd_indices, dim = 2)
 
@@ -152,6 +224,16 @@ class EGNN(nn.Module):
 
         m_ij = self.edge_mlp(edge_input)
 
+        if exists(mask):
+            mask_i = rearrange(mask, 'b i -> b i ()')
+
+            if use_nearest:
+                mask_j = batched_index_select(mask, nbhd_indices, dim = 1)
+                mask = (mask_i * mask_j) & nbhd_mask
+            else:
+                mask_j = rearrange(mask, 'b j -> b () j')
+                mask = mask_i * mask_j
+
         if exists(self.coors_mlp):
             coor_weights = self.coors_mlp(m_ij)
             coor_weights = rearrange(coor_weights, 'b i j () -> b i j')
@@ -159,18 +241,9 @@ class EGNN(nn.Module):
             if self.norm_coor_weights:
                 coor_weights = coor_weights.tanh()
 
-            if self.norm_rel_coors:
-                rel_coors = F.normalize(rel_coors, dim = -1) * self.rel_coors_scale
+            rel_coors = self.coors_norm(rel_coors)
 
             if exists(mask):
-                mask_i = rearrange(mask, 'b i -> b i ()')
-
-                if use_nearest:
-                    mask_j = batched_index_select(mask, nbhd_indices, dim = 1)
-                else:
-                    mask_j = rearrange(mask, 'b j -> b () j')
-
-                mask = mask_i * mask_j
                 coor_weights.masked_fill_(~mask, 0.)
 
             coors_out = einsum('b i j, b i j c -> b i c', coor_weights, rel_coors) + coors
@@ -178,7 +251,19 @@ class EGNN(nn.Module):
             coors_out = coors
 
         if exists(self.node_mlp):
-            m_i = m_ij.sum(dim = -2)
+            m_ij_mask = rearrange(mask, '... -> ... ()')
+            m_ij = m_ij.masked_fill(~m_ij_mask, 0.)
+
+            if self.m_pool_method == 'mean':
+                if exists(mask):
+                    # masked mean
+                    mask_sum = m_ij_mask.sum(dim = -2)
+                    m_i = safe_div(m_ij.sum(dim = -2), mask_sum)
+                else:
+                    m_i = m_ij.mean(dim = -2)
+
+            elif self.m_pool_method == 'sum':
+                m_i = m_ij.sum(dim = -2)
 
             normed_feats = self.node_norm(feats)
             node_mlp_input = torch.cat((normed_feats, m_i), dim = -1)
@@ -188,6 +273,36 @@ class EGNN(nn.Module):
 
         return node_out, coors_out
 
+class EGNN_Network(nn.Module):
+    def __init__(
+        self,
+        *,
+        depth,
+        dim,
+        num_tokens = None,
+        num_edge_tokens = None,
+        edge_dim = 0,
+        **kwargs
+    ):
+        super().__init__()
+        self.token_emb = nn.Embedding(num_tokens, dim) if exists(num_tokens) else None
+        self.edge_emb = nn.Embedding(num_edge_tokens, edge_dim) if exists(num_edge_tokens) else None
+
+        self.layers = nn.ModuleList([])
+        for _ in range(depth):
+            self.layers.append(EGNN(dim = dim, edge_dim = edge_dim, norm_feats = True, **kwargs))
+
+    def forward(self, feats, coors, adj_mat = None, edges = None, mask = None):
+        if exists(self.token_emb):
+            feats = self.token_emb(feats)
+
+        if exists(edges) and exists(self.edge_emb):
+            edges = self.edge_emb(edges)
+
+        for layer in self.layers:
+            feats, coors = layer(feats, coors, adj_mat = adj_mat, edges = edges, mask = mask)
+
+        return feats, coors
 
 class EGNN_sparse(MessagePassing):
     """ Different from the above since it separates the edge assignment
@@ -201,19 +316,27 @@ class EGNN_sparse(MessagePassing):
         edge_attr_dim = 0,
         m_dim = 16,
         fourier_features = 0,
+        update_coors = True, 
+        update_feats = True,
+        norm_feats = False,
         norm_rel_coors = False,
         norm_coor_weights = False,
         dropout = 0.,
         init_eps = 1e-3
     ):
         super().__init__()
+        # model params
         self.fourier_features = fourier_features
         self.feats_dim = feats_dim
         self.pos_dim = pos_dim
+        self.norm_feats = norm_feats
+        self.update_coors = update_coors
+        self.update_feats = update_feats
 
         edge_input_dim = (fourier_features * 2) + (feats_dim * 2) + edge_attr_dim + 1
         dropout = nn.Dropout(dropout) if dropout > 0 else nn.Identity()
 
+        # EDGES
         self.edge_mlp = nn.Sequential(
             nn.Linear(edge_input_dim, edge_input_dim * 2),
             dropout,
@@ -222,13 +345,17 @@ class EGNN_sparse(MessagePassing):
             SiLU()
         )
 
+        # NODES
+        self.node_norm = nn.LayerNorm(feats_dim) if norm_feats else nn.Identity()
+
         self.node_mlp = nn.Sequential(
             nn.Linear(feats_dim + m_dim, feats_dim * 2),
             dropout,
             SiLU(),
             nn.Linear(feats_dim * 2, feats_dim),
-        )
+        ) if update_feats else None
 
+        # COORS
         self.norm_coor_weights = norm_coor_weights
         self.norm_rel_coors = norm_rel_coors
         if norm_rel_coors:
@@ -240,7 +367,7 @@ class EGNN_sparse(MessagePassing):
             dropout,
             SiLU(),
             nn.Linear(m_dim * 4, 1)
-        )
+        ) if update_coors else None
         # seems to be needed to keep the network from exploding to NaN with greater depths
         self.init_eps = init_eps
         self.apply(self.init_)
@@ -277,8 +404,7 @@ class EGNN_sparse(MessagePassing):
 
     def message(self, x_i, x_j, edge_attr) -> Tensor:
         m_ij = self.edge_mlp( torch.cat([x_i, x_j, edge_attr], dim=-1) )
-        coor_w = self.coors_mlp(m_ij)
-        return m_ij, coor_w
+        return m_ij
 
     def propagate(self, edge_index: Adj, size: Size = None, **kwargs):
         """The initial call to start propagating messages.
@@ -294,29 +420,42 @@ class EGNN_sparse(MessagePassing):
         coll_dict = self.__collect__(self.__user_args__,
                                      edge_index, size, kwargs)
         msg_kwargs = self.inspector.distribute('message', coll_dict)
-        m_ij, coor_wij = self.message(**msg_kwargs)
-
-        # normalize if needed
-        if self.norm_rel_coors:
-            kwargs["rel_coors"] = F.normalize(kwargs["rel_coors"], dim = -1) * self.rel_coors_scale
-
-        if self.norm_coor_weights:
-            coor_wij = coor_wij.tanh()
-
-        # aggregate them
         aggr_kwargs = self.inspector.distribute('aggregate', coll_dict)
-        m_i     = self.aggregate(m_ij, **aggr_kwargs)
-        mhat_i  = self.aggregate(coor_wij * kwargs["rel_coors"], **aggr_kwargs)
+        update_kwargs = self.inspector.distribute('update', coll_dict)
+        
+        # get messages
+        m_ij = self.message(**msg_kwargs)
 
-        # update node and coors
-        node, coors = kwargs["x"], kwargs["coors"]
-        coors_out  = coors + mhat_i
+        # update coors if specified
+        if self.update_coors:
+            coor_wij = self.coors_mlp(m_ij)
 
-        hidden_out = self.node_mlp( torch.cat([node, m_i], dim = -1) )
-        hidden_out = hidden_out + node
+            # normalize if needed
+            if self.norm_rel_coors:
+                kwargs["rel_coors"] = F.normalize(kwargs["rel_coors"], dim = -1) * self.rel_coors_scale
+
+            if self.norm_coor_weights:
+                coor_wij = coor_wij.tanh()
+
+            mhat_i = self.aggregate(coor_wij * kwargs["rel_coors"], **aggr_kwargs)
+            coors_out = kwargs["coors"] + mhat_i
+        else:
+            coors_out = kwargs["coors"]
+
+        # update feats if specified
+        if self.update_feats:
+            if self.norm_feats:
+                hidden_feats = self.node_norm(kwargs["x"])
+            else:
+                hidden_feats = kwargs["x"]
+
+            m_i = self.aggregate(m_ij, **aggr_kwargs)
+            hidden_out = self.node_mlp( torch.cat([hidden_feats, m_i], dim = -1) )
+            hidden_out = kwargs["x"] + hidden_out
+        else: 
+            hidden_out = kwargs["x"]
 
         # return tuple
-        update_kwargs = self.inspector.distribute('update', coll_dict)
         return self.update((hidden_out, coors_out), **update_kwargs)
 
     def __repr__(self):
@@ -350,8 +489,9 @@ class EGNN_Sparse_Network(nn.Module):
                        fourier_features = 0,
                        embedding_nums=[], embedding_dims=[],
                        edge_embedding_nums=[], edge_embedding_dims=[],
+                       update_coors=True, update_feats=True, norm_feats=True,
                        norm_rel_coors=False, norm_coor_weights=False,
-                       recalc=0, verbose=False):
+                       recalc=0 ):
         super().__init__()
 
         self.n_layers         = n_layers 
@@ -382,10 +522,12 @@ class EGNN_Sparse_Network(nn.Module):
         self.edge_attr_dim    = edge_attr_dim
         self.m_dim            = m_dim
         self.fourier_features = fourier_features
+        self.norm_feats       = norm_feats
+        self.update_feats     = update_feats
+        self.update_coors     = update_coors
         self.norm_rel_coors   = norm_rel_coors
         self.norm_coor_weights= norm_coor_weights
         self.recalc           = recalc
-        self.verbose          = verbose
         
         # instantiate layers
         for i in range(n_layers):
@@ -394,55 +536,39 @@ class EGNN_Sparse_Network(nn.Module):
                                 edge_attr_dim = edge_attr_dim,
                                 m_dim = m_dim,
                                 fourier_features = fourier_features, 
-                                norm_rel_coors=norm_rel_coors,
-                                norm_coor_weights=norm_coor_weights)
+                                norm_feats = norm_feats,
+                                update_feats = update_feats,
+                                update_coors = update_coors,
+                                norm_rel_coors = norm_rel_coors,
+                                norm_coor_weights = norm_coor_weights)
             self.mpnn_layers.append(layer)
 
     def forward(self, x, edge_index, batch, edge_attr,
                 bsize=None, recalc_edge=None, verbose=0):
-        """ Embedding of inputs when necessary, then pass layers.
-            Recalculate edge features every time with the
+        """ Recalculate edge features every `self.recalc_edge` with the
             `recalc_edge` function if self.recalc_edge is set.
+
+            * x: (N, pos_dim+feats_dim) will be unpacked into coors, feats.
         """
-        original_x = x.clone()
-        original_edge_index = edge_index.clone()
-        original_edge_attr = edge_attr.clone()
-        # pick to embedd. embedd sequentially and add to input - points:
-        stop_concat = -len(self.embedding_dims)
-        to_embedd = x[:, stop_concat:].long()
-        for i,emb_layer in enumerate(self.emb_layers):
-            # the portion corresponding to `to_embedd` part gets dropped
-            # at first iter
-            x = torch.cat([ x[:, :stop_concat], 
-                            emb_layer( to_embedd[:, i] ) 
-                          ], dim=-1)
-            stop_concat = x.shape[-1]
-        # pass layers
+        # NODES - Embedd each dim to its target dimensions:
+        x = embedd_token(x, self.embedding_dims, self.emb_layers)
+
+        # regulates wether to embedd edges each layer
+        edges_need_embedding = True  
         for i,layer in enumerate(self.mpnn_layers):
-            # embedd edge items (needed everytime since edge_attr and idxs
-            # are recalculated every pass)
-            stop_concat = -len(self.edge_embedding_dims)
-            to_embedd = edge_attr[:, stop_concat:].long()
-            for j,edge_emb_layer in enumerate(self.edge_emb_layers):
-                # the portion corresponding to `to_embedd` part gets dropped
-                # at first iter
-                edge_attr = torch.cat([ edge_attr[:, :-len(self.edge_embedding_dims) + j], 
-                                        edge_emb_layer( to_embedd[:, j] ) 
-                              ], dim=-1)
-                stop_concat = x.shape[-1]
+            
+            # EDGES - Embedd each dim to its target dimensions:
+            if edges_need_embedding:
+                edge_attr = embedd_token(x, self.edge_embedding_dims, self.edge_emb_layers)
+                edges_need_embedding = False
+
             # pass layers
             x = layer(x, edge_index, edge_attr, size=bsize)
 
             # recalculate edge info - not needed if last layer
             if self.recalc and ((i%self.recalc == 0) and not (i == len(self.mpnn_layers)-1)) :
                 edge_index, edge_attr, _ = recalc_edge(x) # returns attr, idx, any_other_info
-            else: 
-                edge_attr = original_edge_attr.clone()
-                edge_index = original_edge_index.clone()
-            
-            if verbose:
-                print("========")
-                print(i, "layer, nlinks:", edge_attr.shape)
+                edges_need_embedding = True
             
         return x
 
